'use strict';

var reactiveVar = require('meteor/reactive-var');
var templating = require('meteor/templating');
var tracker = require('meteor/tracker');
var meteor = require('meteor/meteor');
var moment_ = require('moment');

function escapeRegexChars(text) {
    return text.replace(/[.^$*+\-?()[\]{}\\|]/g, "\\$&");
}

var KeyValueEntry = /** @class */ (function () {
    function KeyValueEntry(type) {
        this._value = undefined;
        this._type = type;
    }
    Object.defineProperty(KeyValueEntry.prototype, "type", {
        get: function () {
            return this._type;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueEntry.prototype, "id", {
        get: function () {
            return this._type.id;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueEntry.prototype, "label", {
        get: function () {
            return this._type.label;
        },
        enumerable: true,
        configurable: true
    });
    KeyValueEntry.prototype.getSuggestions = function (prefix) {
        return this._type.getSuggestions(prefix);
    };
    KeyValueEntry.prototype.parse = function (val) {
        if (typeof val === "string") {
            return this._type.parseString(val);
        }
        else if (val !== undefined) {
            return this._type.checkValue(val);
        }
        return undefined;
    };
    KeyValueEntry.prototype.set = function (val) {
        this._value = this.parse(val);
        return this._value !== undefined;
    };
    Object.defineProperty(KeyValueEntry.prototype, "value", {
        get: function () {
            return this._value;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueEntry.prototype, "isPartial", {
        get: function () {
            return this.value === undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueEntry.prototype, "display", {
        get: function () {
            if (this.value === undefined) {
                return undefined;
            }
            return this._type.display(this.value);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueEntry.prototype, "editText", {
        get: function () {
            if (this.value === undefined) {
                return undefined;
            }
            return this._type.editText(this.value);
        },
        enumerable: true,
        configurable: true
    });
    KeyValueEntry.prototype.equals = function (entry) {
        return this._type.isEqual(this, entry);
    };
    return KeyValueEntry;
}());

var KeyValueType = /** @class */ (function () {
    function KeyValueType(id, label, aliases) {
        if (aliases === void 0) { aliases = []; }
        this._id = id;
        this._label = label;
        this._aliases = [id, label].concat(aliases).map(function (a) { return a.toLowerCase(); });
    }
    Object.defineProperty(KeyValueType.prototype, "id", {
        get: function () {
            return this._id;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueType.prototype, "label", {
        get: function () {
            return this._label;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(KeyValueType.prototype, "aliases", {
        get: function () {
            return this._aliases;
        },
        enumerable: true,
        configurable: true
    });
    KeyValueType.prototype.findAlias = function (substring) {
        var substringLower = substring.toLowerCase();
        return this.aliases.find(function (a) { return a.includes(substringLower); });
    };
    KeyValueType.prototype.createPartialEntry = function () {
        return new KeyValueEntry(this);
    };
    KeyValueType.prototype.tryCreatePartialEntry = function (key) {
        if (!this.aliases.includes(key.toLowerCase())) {
            return undefined;
        }
        return new KeyValueEntry(this);
    };
    KeyValueType.prototype.tryCreateFullEntry = function (val) {
        var entry = new KeyValueEntry(this);
        if (entry.set(val)) {
            return entry;
        }
        else {
            return undefined;
        }
    };
    KeyValueType.prototype.isEqual = function (entryA, entryB) {
        return entryA.type === entryB.type && entryA.value === entryB.value;
    };
    return KeyValueType;
}());

var __extends = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var trueValues = ["ja", "yes", "wahr", "true", "1", "✓"];
var falseValues = ["nein", "no", "falsch", "false", "0", "✗"];
var BooleanKeyValueType = /** @class */ (function (_super) {
    __extends(BooleanKeyValueType, _super);
    function BooleanKeyValueType() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    BooleanKeyValueType.validate = function (val) {
        if (typeof val === "boolean") {
            return val;
        }
        else {
            throw new meteor.Meteor.Error("invalid-boolean", "'" + val + "' is not a valid boolean.");
        }
    };
    BooleanKeyValueType.prototype.getSuggestions = function (prefix) {
        var substringLower = prefix.toLowerCase();
        var trueValue = trueValues.find(function (val) { return val.includes(substringLower); });
        var falseValue = falseValues.find(function (val) { return val.includes(substringLower); });
        return [
            { display: this.display(true), match: trueValue, value: trueValues[0] },
            { display: this.display(false), match: falseValue, value: falseValues[0] },
        ].filter(function (o) { return o.match !== undefined; });
    };
    BooleanKeyValueType.prototype.parseString = function (val) {
        var value = val.toLowerCase();
        if (value === "" || trueValues.includes(value)) {
            return true;
        }
        else if (falseValues.includes(value)) {
            return false;
        }
        else {
            return undefined;
        }
    };
    BooleanKeyValueType.prototype.checkValue = function (val) {
        if (typeof val === "boolean") {
            return val;
        }
        else {
            return undefined;
        }
    };
    BooleanKeyValueType.prototype.display = function (value) {
        if (value) {
            return { text: "✓" };
        }
        else {
            return { text: "✗" };
        }
    };
    BooleanKeyValueType.prototype.editText = function (value) {
        if (value) {
            return trueValues[0];
        }
        else {
            return falseValues[0];
        }
    };
    return BooleanKeyValueType;
}(KeyValueType));

var __extends$1 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var moment = moment_;
var symbolForOperator = {
    "<": "<",
    ">": ">",
    "<=": "≤",
    ">=": "≥",
    "!=": "≠",
    "=": "=",
};
var operatorsAndSymbols = ["<", ">", "<=", ">=", "!=", "=", "≤", "≥", "≠"];
function toSymbol(op) {
    // @ts-ignore missing index signature of mapped type
    return symbolForOperator[op] || op;
}
var selectorForOperator = {
    "<": "$lt",
    ">": "$gt",
    "<=": "$lte",
    ">=": "$gte",
    "!=": "$ne",
    "=": "$eq",
};
function toSelector(op) {
    // @ts-ignore missing index signature of mapped type
    return selectorForOperator[op];
}
var DateKeyValueType = /** @class */ (function (_super) {
    __extends$1(DateKeyValueType, _super);
    function DateKeyValueType() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._dateFormat = "YYYY-MM-DD";
        _this._dateParseFormats = ["YYYY-MM-DD", "YY-MM-DD", "MM-DD"];
        return _this;
    }
    DateKeyValueType.mongoSelectorFor = function (value) {
        var _a;
        return _a = {}, _a[toSelector(value.operator)] = value.date, _a;
    };
    Object.defineProperty(DateKeyValueType.prototype, "dateFormat", {
        set: function (dateFormat) {
            this._dateFormat = dateFormat;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DateKeyValueType.prototype, "dateParseFormats", {
        set: function (dateParseFormats) {
            this._dateParseFormats = dateParseFormats;
        },
        enumerable: true,
        configurable: true
    });
    DateKeyValueType.prototype.getSuggestions = function (prefix) {
        var _this = this;
        var parsed = this.parseString(prefix);
        // empty input or invalid input or fallback "="
        if (prefix === "" || !parsed || (parsed.operator === "=" && !prefix.startsWith("="))) {
            var date_1 = new Date();
            var operators = Object.keys(symbolForOperator);
            if (prefix !== "") {
                date_1 = parsed ? parsed.date : date_1;
                var filteredOperators = operatorsAndSymbols.filter(function (op) { return op.startsWith(prefix); });
                operators = filteredOperators.length > 0 ? filteredOperators : operators;
            }
            return operators.map(function (operator) { return ({
                display: _this.display({ operator: operator, date: date_1 }),
                value: _this.editText({ operator: operator, date: date_1 }),
            }); });
        }
        else {
            return [];
        }
    };
    DateKeyValueType.prototype.parseString = function (val) {
        var value = val.trim();
        if (value === "") {
            value = "=" + new Date();
        }
        value = value.replace(/\s/g, ""); // remove spaces
        var match = value.replace("≥", ">=")
            .replace("≤", "<=")
            .replace("<>", "!=")
            .replace("≠", "!=")
            .replace("==", "=")
            .match(/^(<=|<|>=|>|!=|=)?(.+)/);
        if (match) {
            var parsedDate = moment.utc(match[2], this._dateParseFormats, true);
            if (parsedDate.isValid()) {
                return {
                    operator: match[1] || "=",
                    date: parsedDate.toDate(),
                };
            }
            else {
                return undefined;
            }
        }
        else {
            return undefined;
        }
    };
    DateKeyValueType.prototype.checkValue = function (val) {
        if (Object.keys(symbolForOperator).includes(val.operator) && val.date instanceof Date && !isNaN(val.date)) {
            return val;
        }
        else {
            return undefined;
        }
    };
    DateKeyValueType.prototype.display = function (value) {
        return { text: toSymbol(value.operator) + " " + moment(value.date).format(this._dateFormat) };
    };
    DateKeyValueType.prototype.editText = function (value) {
        return this.display(value).text;
    };
    return DateKeyValueType;
}(KeyValueType));

var __extends$2 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var symbolForOperator$1 = {
    "<": "<",
    ">": ">",
    "<=": "≤",
    ">=": "≥",
    "!=": "≠",
    "=": "=",
};
var operatorsAndSymbols$1 = ["<", ">", "<=", ">=", "!=", "=", "≤", "≥", "≠"];
function toSymbol$1(op) {
    // @ts-ignore missing index signature of mapped type
    return symbolForOperator$1[op] || op;
}
var selectorForOperator$1 = {
    "<": "$lt",
    ">": "$gt",
    "<=": "$lte",
    ">=": "$gte",
    "!=": "$ne",
    "=": "$eq",
};
function toSelector$1(op) {
    // @ts-ignore missing index signature of mapped type
    return selectorForOperator$1[op];
}
var NumberKeyValueType = /** @class */ (function (_super) {
    __extends$2(NumberKeyValueType, _super);
    function NumberKeyValueType() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._delimiters = {
            decimal: ".",
            groups: ",",
        };
        _this._groupsDelimiterRegex = /,/gi;
        return _this;
    }
    NumberKeyValueType.mongoSelectorFor = function (value, numberTransform) {
        if (numberTransform === void 0) { numberTransform = function (n) { return n; }; }
        var _a;
        return _a = {}, _a[toSelector$1(value.operator)] = numberTransform(value.number), _a;
    };
    Object.defineProperty(NumberKeyValueType.prototype, "delimiters", {
        set: function (delimiters) {
            this._delimiters = delimiters;
            this._groupsDelimiterRegex = new RegExp(escapeRegexChars(this._delimiters.groups), "gi");
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NumberKeyValueType.prototype, "numberFormatter", {
        set: function (numberFormatter) {
            this._numberFormatter = numberFormatter;
        },
        enumerable: true,
        configurable: true
    });
    NumberKeyValueType.prototype.getSuggestions = function (prefix) {
        var _this = this;
        var parsed = this.parseString(prefix);
        // empty input or invalid input or fallback "="
        if (prefix === "" || !parsed || (parsed.operator === "=" && !prefix.startsWith("="))) {
            var number_1 = 42;
            var operators = Object.keys(symbolForOperator$1);
            if (prefix !== "") {
                number_1 = parsed ? parsed.number : number_1;
                var filteredOperators = operatorsAndSymbols$1.filter(function (op) { return op.startsWith(prefix); });
                operators = filteredOperators.length > 0 ? filteredOperators : operators;
            }
            return operators.map(function (operator) { return ({
                display: _this.display({ operator: operator, number: number_1 }),
                value: _this.editText({ operator: operator, number: number_1 }),
            }); });
        }
        else {
            return [];
        }
    };
    NumberKeyValueType.prototype.parseString = function (val) {
        var value = val.trim();
        if (value === "") {
            value = ">0";
        }
        value = value.replace(/\s/g, "") // remove spaces
            .replace(this._groupsDelimiterRegex, "") // remove grouping delimiters
            .replace(this._delimiters.decimal, "."); // convert to decimal point
        var match = value.replace("≥", ">=")
            .replace("≤", "<=")
            .replace("<>", "!=")
            .replace("≠", "!=")
            .replace("==", "=")
            .match(/^(<=|<|>=|>|!=|=)?(-?\d*\.?\d+)/);
        if (match) {
            return {
                operator: match[1] || "=",
                number: Number(match[2]),
            };
        }
        else {
            return undefined;
        }
    };
    NumberKeyValueType.prototype.checkValue = function (val) {
        if (Object.keys(symbolForOperator$1).includes(val.operator) && typeof val.number === "number") {
            return val;
        }
        else {
            return undefined;
        }
    };
    NumberKeyValueType.prototype.display = function (value) {
        var numberString = this._numberFormatter ? this._numberFormatter.format(value.number) : String(value.number);
        return { text: toSymbol$1(value.operator) + " " + numberString };
    };
    NumberKeyValueType.prototype.editText = function (value) {
        return this.display(value).text;
    };
    return NumberKeyValueType;
}(KeyValueType));

var __extends$3 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var symbolForOperator$2 = {
    "=": "=",
    "*": "*",
    "/": "÷",
    "+": "+",
    "-": "-",
};
var operatorsAndSymbols$2 = ["=", "*", "/", "+", "-", "÷"];
function toSymbol$2(op) {
    // @ts-ignore missing index signature of mapped type
    return symbolForOperator$2[op] || op;
}
var NumberModifierKeyValueType = /** @class */ (function (_super) {
    __extends$3(NumberModifierKeyValueType, _super);
    function NumberModifierKeyValueType() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._delimiters = {
            decimal: ".",
            groups: ",",
        };
        _this._groupsDelimiterRegex = /,/gi;
        return _this;
    }
    NumberModifierKeyValueType.apply = function (value, number, roundMode) {
        if (roundMode === void 0) { roundMode = "not"; }
        var result;
        switch (value.operator) {
            case "=":
                result = value.number;
                break;
            case "*":
                result = number * value.number;
                break;
            case "/":
                result = number / value.number;
                break;
            case "+":
                result = number + value.number;
                break;
            case "-":
                result = number - value.number;
                break;
            default:
                throw new meteor.Meteor.Error("invalid-number-modifier", "'" + value.operator + "' is not a valid operation on numbers.");
        }
        switch (roundMode) {
            case "up":
                result = Math.ceil(result);
                break;
            case "down":
                result = Math.floor(result);
                break;
            case "math":
                result = Math.round(result);
                break;
            case "not":
            default:
            // no rounding
        }
        return result;
    };
    Object.defineProperty(NumberModifierKeyValueType.prototype, "delimiters", {
        set: function (delimiters) {
            this._delimiters = delimiters;
            this._groupsDelimiterRegex = new RegExp(escapeRegexChars(this._delimiters.groups), "gi");
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(NumberModifierKeyValueType.prototype, "numberFormatter", {
        set: function (numberFormatter) {
            this._numberFormatter = numberFormatter;
        },
        enumerable: true,
        configurable: true
    });
    NumberModifierKeyValueType.prototype.getSuggestions = function (prefix) {
        var _this = this;
        var parsed = this.parseString(prefix);
        // empty input or invalid input or fallback "="
        if (prefix === "" || !parsed || (parsed.operator === "=" && !prefix.startsWith("="))) {
            var number_1 = 42;
            var operators = Object.keys(symbolForOperator$2);
            if (prefix !== "") {
                number_1 = parsed ? parsed.number : number_1;
                var filteredOperators = operatorsAndSymbols$2.filter(function (op) { return op.startsWith(prefix); });
                operators = filteredOperators.length > 0 ? filteredOperators : operators;
            }
            return operators.map(function (operator) { return ({
                display: _this.display({ operator: operator, number: number_1 }),
                value: _this.editText({ operator: operator, number: number_1 }),
            }); });
        }
        else {
            return [];
        }
    };
    NumberModifierKeyValueType.prototype.parseString = function (val) {
        var value = val.trim();
        if (value === "") {
            return undefined;
        }
        value = value.replace(/\s/g, "") // remove spaces
            .replace(this._groupsDelimiterRegex, "") // remove grouping delimiters
            .replace(this._delimiters.decimal, "."); // convert to decimal point
        var match = value.replace("÷", "/")
            .match(/^([=*/+-])?(-?\d*\.?\d+)/);
        if (match) {
            return {
                operator: match[1] || "=",
                number: Number(match[2]),
            };
        }
        else {
            return undefined;
        }
    };
    NumberModifierKeyValueType.prototype.checkValue = function (val) {
        if (Object.keys(symbolForOperator$2).includes(val.operator) && typeof val.number === "number") {
            return val;
        }
        else {
            return undefined;
        }
    };
    NumberModifierKeyValueType.prototype.display = function (value) {
        var numberString = this._numberFormatter ? this._numberFormatter.format(value.number) : String(value.number);
        return { text: toSymbol$2(value.operator) + " " + numberString };
    };
    NumberModifierKeyValueType.prototype.editText = function (value) {
        return this.display(value).text;
    };
    return NumberModifierKeyValueType;
}(KeyValueType));

var __extends$4 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var RegexKeyValueType = /** @class */ (function (_super) {
    __extends$4(RegexKeyValueType, _super);
    function RegexKeyValueType() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    RegexKeyValueType.mongoSelectorFor = function (value) {
        return { $regex: value.regex, $options: value.options };
    };
    RegexKeyValueType.prototype.getSuggestions = function (prefix) {
        var _this = this;
        var display = function (val) { return _this.display(_this.parseString(val)); };
        return [
            { display: display("abc"), value: "abc", extra: "Einfache Suche" },
            { display: display('"Hallo Welt!"'), value: '"Hallo Welt!"', extra: "\"...\" für Leerzeichen" },
            { display: display("/[a-c0-9]/i"), value: "/[a-c0-9]/i", extra: "RegEx" },
        ];
    };
    RegexKeyValueType.prototype.parseString = function (val) {
        var text = val.trim();
        if (text === "") {
            return undefined;
        }
        var match = text.match(/^\/(.*)\/([a-z]*)$/);
        if (match) {
            try {
                // tslint:disable-next-line:no-unused-expression as the expression is used for error checking
                new RegExp(match[1], match[2]);
                return {
                    text: text,
                    printRegex: jQuery("<div>").text(match[1]).html(),
                    regex: match[1],
                    options: match[2],
                };
            }
            catch (e) {
                // escape below like non-regex strings
            }
        }
        return {
            text: text,
            regex: escapeRegexChars(text),
            options: "i",
        };
    };
    RegexKeyValueType.prototype.checkValue = function (val) {
        if (typeof val.text === "string"
            && (val.printRegex === undefined || typeof val.printRegex === "string")
            && typeof val.regex === "string"
            && typeof val.options === "string") {
            return val;
        }
        else {
            return undefined;
        }
    };
    RegexKeyValueType.prototype.display = function (value) {
        if (value.printRegex) {
            return { html: "<code>/</code>" + value.printRegex + "<code>/" + value.options + "</code>" };
        }
        else {
            return { text: value.text };
        }
    };
    RegexKeyValueType.prototype.editText = function (value) {
        if (value.text.includes(" ")) {
            return "\"" + value.text.replace(/\\/g, "\\\\").replace(/"/g, "\\\"") + "\"";
        }
        else {
            return value.text;
        }
    };
    return RegexKeyValueType;
}(KeyValueType));

var __extends$5 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var ascendingValues = ["aufsteigend", "auf", "asc", "ascending", "up", "1", "+"];
var descendingValues = ["absteigend", "ab", "desc", "descending", "down", "-1", "-"];
var SortKeyValueType = /** @class */ (function (_super) {
    __extends$5(SortKeyValueType, _super);
    function SortKeyValueType() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    SortKeyValueType.mongoSpecifierFor = function (value) {
        return value ? "asc" : "desc";
    };
    SortKeyValueType.prototype.getSuggestions = function (prefix) {
        var substringLower = prefix.toLowerCase();
        var ascendingValue = ascendingValues.find(function (val) { return val.includes(substringLower); });
        var descendingValue = descendingValues.find(function (val) { return val.includes(substringLower); });
        return [
            { display: this.display(true), match: ascendingValue, value: ascendingValues[0] },
            { display: this.display(false), match: descendingValue, value: descendingValues[0] },
        ].filter(function (o) { return o.match !== undefined; });
    };
    SortKeyValueType.prototype.parseString = function (val) {
        var value = val.trim().toLowerCase();
        if (value === "" || ascendingValues.includes(value)) {
            return true;
        }
        else if (descendingValues.includes(value)) {
            return false;
        }
        else {
            return undefined;
        }
    };
    SortKeyValueType.prototype.checkValue = function (val) {
        if (typeof val === "boolean") {
            return val;
        }
        else {
            return undefined;
        }
    };
    SortKeyValueType.prototype.display = function (value) {
        if (value) {
            return { html: "<i class='fa fa-sort-up'></i>" };
        }
        else {
            return { html: "<i class='fa fa-sort-down'></i>" };
        }
    };
    SortKeyValueType.prototype.editText = function (value) {
        if (value) {
            return ascendingValues[0];
        }
        else {
            return descendingValues[0];
        }
    };
    return SortKeyValueType;
}(KeyValueType));

var __extends$6 = (undefined && undefined.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var TagKeyValueType = /** @class */ (function (_super) {
    __extends$6(TagKeyValueType, _super);
    function TagKeyValueType() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._tags = [];
        _this._allowOther = false;
        return _this;
    }
    Object.defineProperty(TagKeyValueType.prototype, "tags", {
        set: function (tags) {
            this._tags = tags;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TagKeyValueType.prototype, "allowOther", {
        set: function (allowOther) {
            this._allowOther = allowOther;
        },
        enumerable: true,
        configurable: true
    });
    TagKeyValueType.prototype.getSuggestions = function (prefix) {
        var substringLower = prefix.toLowerCase();
        return this._tags.filter(function (t) { return t.toLowerCase().includes(substringLower); })
            .map(function (t) { return ({
            display: { text: t },
            match: prefix !== "" ? t : undefined,
            value: t,
        }); });
    };
    TagKeyValueType.prototype.parseString = function (val) {
        if (this._allowOther && val.trim() !== "") {
            return val.trim();
        }
        var value = val.trim().toLowerCase();
        return this._tags.find(function (t) { return t.toLowerCase() === value; });
    };
    TagKeyValueType.prototype.checkValue = function (val) {
        if (typeof val === "string") {
            return this.parseString(val);
        }
        else {
            return undefined;
        }
    };
    TagKeyValueType.prototype.display = function (value) {
        return { text: value };
    };
    TagKeyValueType.prototype.editText = function (value) {
        return value;
    };
    return TagKeyValueType;
}(KeyValueType));

function createEntriesFromTypes(entries, types) {
    return entries.map(function (e) { return createEntry(e, types); })
        .filter(function (entry) { return entry !== undefined; });
}
function createEntry(e, types) {
    for (var _i = 0, types_1 = types; _i < types_1.length; _i++) {
        var type = types_1[_i];
        if (type.id === e.key) {
            return type.tryCreateFullEntry(e.value);
        }
    }
    return undefined;
}

function KeyValueInputTemplateData() {
    return templating.Template.currentData();
}
function KeyValueInputTemplate() {
    return templating.Template.instance();
}
templating.Template.keyValueInput.onCreated(function () {
    var _this = this;
    this.autorun(function () {
        var data = KeyValueInputTemplateData();
        data.kind = data.kind || "multiple";
    });
    this.entries = new reactiveVar.ReactiveVar(createEntries(this.data.entries, this));
    this.partialEntry = new reactiveVar.ReactiveVar(null);
    resetPartialEntry(this);
    this.defaultEntries = new reactiveVar.ReactiveVar([]);
    this.autorun(function () { return createDefaultEntries(_this); });
    this.suggestions = new reactiveVar.ReactiveVar(null);
    this.autorun(function () { return buildSuggestions(_this); });
});
templating.Template.keyValueInput.onRendered(function () {
    this.textInput = this.$(".key-value-text-input");
    this.element = this.$(".key-value-input")[0];
});
templating.Template.keyValueInput.helpers({
    isSingle: function () {
        return KeyValueInputTemplate().data.kind === "single";
    },
    defaultEntries: function () {
        return KeyValueInputTemplate().defaultEntries.get();
    },
    entries: function () {
        return KeyValueInputTemplate().entries.get();
    },
    partialEntry: function () {
        return KeyValueInputTemplate().partialEntry.get();
    },
    availableTypes: function () {
        var data = KeyValueInputTemplateData();
        return data.kind === "multiple" ? data.types.map(function (t) { return t.label; }).join(": / ") + ":" : "";
    },
    suggestions: function () {
        return KeyValueInputTemplate().suggestions.get();
    },
});
templating.Template.keyValueInput.events({
    "click .key-value-input": function (event, templateInstance) {
        moveFocus(event, templateInstance, "text-input");
    },
    "input .key-value-input, focus .key-value-input": function (event, templateInstance) {
        buildSuggestions(templateInstance);
    },
    "keydown .key-value-text-input": function (event, templateInstance) {
        var key = event.key;
        var target = event.target;
        if (key === ":") {
            if (addPartialEntry(templateInstance.textInput.val(), templateInstance)) {
                event.preventDefault();
                templateInstance.textInput.val("");
                buildSuggestions(templateInstance);
            }
        }
        else if (key === "Enter" || key === "Tab") {
            if (templateInstance.textInput.val() !== "" || templateInstance.partialEntry.get()) {
                event.preventDefault();
                if (completeEntry(templateInstance.textInput.val(), templateInstance)) {
                    templateInstance.textInput.val("");
                    buildSuggestions(templateInstance);
                }
            }
        }
        else if (key === "Backspace" && target.selectionStart === 0 && target.selectionEnd === 0) {
            var prependText = editLastEntry(templateInstance);
            templateInstance.textInput.val(prependText + templateInstance.textInput.val());
            target.setSelectionRange(prependText.length, prependText.length);
            buildSuggestions(templateInstance);
            event.preventDefault();
        }
        else if (key === "Escape") {
            templateInstance.suggestions.set(null);
        }
    },
    "keyup .key-value-text-input": function (event, templateInstance) {
        var key = event.key;
        var target = event.target;
        if (key === " ") {
            var inputValue = templateInstance.textInput.val();
            var cursorPos = target.selectionStart;
            var value = inputValue.trim();
            if (cursorPos === inputValue.length) {
                if (!value.startsWith("\"") || value.match(/^"(?:\\.|[^\\])*"$/)) {
                    if (completeEntry(value, templateInstance)) {
                        templateInstance.textInput.val("");
                        buildSuggestions(templateInstance);
                    }
                }
            }
            else if (!value.startsWith("\"")) {
                // shift cursor due to trimmed whitespace and escaped characters in front of the cursor position
                var shiftCursor = inputValue.length - inputValue.trimLeft().length + 1
                    + (inputValue.substring(0, cursorPos).match(/["\\]/g) || []).length;
                // escape \ and "
                templateInstance.textInput.val("\"" + value.replace(/\\/g, "\\\\").replace(/"/g, "\\\"") + "\"");
                target.setSelectionRange(cursorPos + shiftCursor, cursorPos + shiftCursor);
                buildSuggestions(templateInstance);
            }
        }
        else if (key === "ArrowDown" || (key === "ArrowLeft" && target.selectionEnd === 0)) {
            event.preventDefault();
            moveFocus(event, templateInstance, key);
        }
    },
    "click .key-value-suggestion": function (event, templateInstance) {
        event.preventDefault();
        if ((!templateInstance.partialEntry.get() && addPartialEntry(this.suggestion.value, templateInstance))
            || (completeEntry(this.suggestion.value, templateInstance))) {
            templateInstance.textInput.val("");
            buildSuggestions(templateInstance);
        }
    },
    "keydown .key-value-suggestion": function (event, templateInstance) {
        var key = event.key;
        var target = event.currentTarget;
        if (["ArrowUp", "ArrowDown", "Home", "End"].includes(key)) {
            event.preventDefault();
            event.stopPropagation();
            moveFocus(event, templateInstance, key);
        }
        else if (["Enter", ":", " "].includes(key)) {
            event.preventDefault();
            templateInstance.$(target).click();
        }
        else if (key.length === 1 || ["Backspace", "Delete", "ArrowLeft", "ArrowRight"].includes(key)) {
            moveFocus(event, templateInstance, "text-input");
        }
    },
    "keydown .key-value-entry": function (event, templateInstance) {
        var key = event.key;
        if (["ArrowLeft", "ArrowRight"].includes(key)) {
            moveFocus(event, templateInstance, key);
        }
    },
    "keyup .key-value-entry": function (event, templateInstance) {
        var key = event.key;
        if (key === "Backspace" || key === "Delete") {
            event.preventDefault();
            if (!this.entry.isPartial) {
                removeEntry(this.entry, templateInstance);
                moveFocus(event, templateInstance, this.index); // index should always be set for non-partial entries
            }
            else if (templateInstance.data.kind !== "single") { // single type always needs a partial entry
                templateInstance.partialEntry.set(null);
                moveFocus(event, templateInstance, "ArrowRight");
            }
        }
    },
    "click .key-value-entry": function (event) {
        event.stopPropagation();
        event.currentTarget.focus();
    },
    "click .key-value-entry .fa-times": function (event, templateInstance) {
        event.stopPropagation();
        removeEntry(this.entry, templateInstance);
        moveFocus(event, templateInstance, "text-input");
    },
    "click .fa-times.remove-all": function (event, templateInstance) {
        event.stopPropagation();
        removeAllEntries(templateInstance);
        moveFocus(event, templateInstance, "text-input");
    },
    "clearEntries": function (event, templateInstance) {
        removeAllEntries(templateInstance);
    },
});
function moveFocus(event, templateInstance, to) {
    // Set focus
    if (to === "text-input") {
        templateInstance.textInput[0].focus();
        return;
    }
    else if (typeof to === "number") { // entry
        tracker.Tracker.afterFlush(function () {
            var siblings = templateInstance.$(".key-value-entry, .key-value-text-input");
            if (siblings[to + templateInstance.defaultEntries.get().length]) {
                siblings[to + templateInstance.defaultEntries.get().length].focus();
            }
        });
        return;
    }
    // Move Focus
    var target;
    var currentTarget = event.currentTarget;
    if (to === "ArrowDown") { // suggestion or text input
        if (currentTarget.classList.contains("key-value-suggestion")) {
            target = templateInstance.$(currentTarget).next(".key-value-suggestion");
        }
        else {
            target = templateInstance.$(".key-value-suggestion");
        }
    }
    else if (to === "ArrowUp") { // suggestion
        target = templateInstance.$(currentTarget).prev(".key-value-suggestion");
        if (!target[0]) {
            target = templateInstance.textInput;
        }
    }
    else if (to === "End") { // suggestion
        target = templateInstance.$(".key-value-suggestion").filter(":last");
    }
    else if (to === "Home") { // suggestion
        target = templateInstance.$(".key-value-suggestion");
    }
    else if (to === "ArrowRight") { // entry
        target = templateInstance.$(currentTarget).next(".key-value-entry");
        if (!target[0]) {
            target = templateInstance.textInput;
        }
    }
    else if (to === "ArrowLeft") { // entry or text input
        if (currentTarget.classList.contains("key-value-entry")) {
            target = templateInstance.$(currentTarget).prev(".key-value-entry");
        }
        else {
            target = templateInstance.$(".key-value-entry").filter(":last");
        }
    }
    if (target[0]) {
        target[0].focus();
    }
}
function resetPartialEntry(templateInstance) {
    if (templateInstance.data.kind === "single") {
        templateInstance.partialEntry.set(templateInstance.data.type.createPartialEntry());
    }
    else {
        templateInstance.partialEntry.set(null);
    }
}
function addPartialEntry(val, templateInstance) {
    if (templateInstance.partialEntry.get() || templateInstance.data.kind === "single") {
        return false;
    }
    var typeName = val.trim().toLowerCase();
    for (var _i = 0, _a = templateInstance.data.types; _i < _a.length; _i++) {
        var type = _a[_i];
        var entry = type.tryCreatePartialEntry(typeName);
        if (entry) {
            templateInstance.partialEntry.set(entry);
            return true;
        }
    }
    return false;
}
function getFallbackType(templateInstance) {
    if (templateInstance.data.kind === "multiple" && templateInstance.data.fallbackType !== undefined) {
        return templateInstance.data.types[templateInstance.data.fallbackType];
    }
    else {
        return undefined;
    }
}
function completeEntry(val, templateInstance) {
    var partialEntry = templateInstance.partialEntry.get();
    var fallbackType = getFallbackType(templateInstance);
    var trimmedValue = val.trim();
    if (trimmedValue.startsWith("\"")) {
        // remove leading and possible trailing ", then replace escaped characters
        trimmedValue = trimmedValue.replace(/^"((?:\\.|[^\\])*?)"?$/g, "$1").replace(/\\(.)/g, "$1").trim();
    }
    var newEntry;
    if (partialEntry) {
        newEntry = partialEntry;
        if (!newEntry.set(trimmedValue)) {
            return false;
        }
        resetPartialEntry(templateInstance);
    }
    else {
        if (fallbackType) {
            newEntry = fallbackType.tryCreateFullEntry(trimmedValue);
        }
        if (!newEntry) {
            // if nothing worked out, try to create a partial entry from the input
            return addPartialEntry(trimmedValue, templateInstance);
        }
    }
    var entries = templateInstance.entries.get();
    if (entries.some(function (e) { return e.equals(newEntry); })) {
        return true;
    }
    entries.push(newEntry);
    templateInstance.entries.dep.changed();
    emitChange(templateInstance);
    return true;
}
function createDefaultEntries(templateInstance) {
    templateInstance.defaultEntries.set(createEntries(KeyValueInputTemplateData().defaultEntries, templateInstance)
        .map(function (entry) { entry.isDefault = true; return entry; }));
}
function createEntries(entries, templateInstance) {
    var types = templateInstance.data.kind === "single" ?
        [templateInstance.data.type] : templateInstance.data.types;
    return createEntriesFromTypes(entries || [], types);
}
function editLastEntry(templateInstance) {
    if (templateInstance.data.kind !== "single") { // cannot edit partial entry of single mode input
        var partialEntry = templateInstance.partialEntry.get();
        if (partialEntry) {
            templateInstance.partialEntry.set(null);
            return partialEntry.label;
        }
    }
    var lastEntry = templateInstance.entries.get().splice(-1, 1)[0];
    if (lastEntry) {
        templateInstance.entries.dep.changed();
        var editValue = lastEntry.editText || "";
        lastEntry.set(undefined);
        templateInstance.partialEntry.set(lastEntry);
        emitChange(templateInstance);
        return editValue;
    }
    return "";
}
function removeEntry(entry, templateInstance) {
    var filtered = templateInstance.entries.get().filter(function (e) { return !e.equals(entry); });
    templateInstance.entries.set(filtered);
    emitChange(templateInstance);
}
function removeAllEntries(templateInstance) {
    templateInstance.entries.set([]);
    resetPartialEntry(templateInstance);
    templateInstance.textInput.val("");
    emitChange(templateInstance);
}
function emitChange(templateInstance) {
    tracker.Tracker.flush();
    if (templateInstance.element) {
        var event_1 = new CustomEvent("keyValueEntriesChanged", { detail: templateInstance.entries.get() });
        tracker.Tracker.afterFlush(function () { return window.setTimeout(function () { return templateInstance.element.dispatchEvent(event_1); }, 100); }); // make sure keyValueInput stays responsive; apply changes a little later
    }
}
function buildSuggestions(templateInstance) {
    var partialEntry = templateInstance.partialEntry.get();
    var inputValue = templateInstance.textInput ? templateInstance.textInput.val().trim() : "";
    var suggestions = null;
    if (partialEntry) {
        var entrySuggestions = partialEntry.getSuggestions(inputValue);
        var parsed = partialEntry.parse(inputValue);
        if (parsed !== undefined) {
            entrySuggestions.push({
                display: partialEntry.type.display(parsed),
                value: partialEntry.type.editText(parsed),
                fallback: true,
            });
        }
        if (entrySuggestions && entrySuggestions.length > 0) {
            suggestions = entrySuggestions;
        }
    }
    else {
        var types = templateInstance.data.types;
        if (inputValue === "") {
            suggestions = types.map(function (t) { return ({ display: { text: t.label }, value: t.id }); });
        }
        else {
            var matchingTypes = types
                .map(function (t) { return ({ display: { text: t.label }, match: t.findAlias(inputValue), value: t.id }); })
                .filter(function (o) { return o.match; });
            if (matchingTypes.length > 0) {
                suggestions = matchingTypes;
            }
            var fallbackType = getFallbackType(templateInstance);
            if (fallbackType) {
                var fallbackSuggestion = {
                    display: { text: fallbackType.label + ": " + inputValue },
                    value: inputValue,
                    fallback: true,
                };
                if (!suggestions) {
                    suggestions = [];
                }
                suggestions.push(fallbackSuggestion);
            }
        }
        if (suggestions) {
            suggestions.header = "Mit <kbd>:</kbd> bestätigen oder aus Liste wählen.";
        }
    }
    highlightMatches(suggestions, inputValue);
    templateInstance.suggestions.set(suggestions);
}
function highlightMatches(suggestions, substring) {
    if (!suggestions) {
        return;
    }
    var escaped = escapeRegexChars(substring);
    for (var _i = 0, suggestions_1 = suggestions; _i < suggestions_1.length; _i++) {
        var suggestion = suggestions_1[_i];
        if (suggestion.match) {
            suggestion.matchHtml = suggestion.match.replace(new RegExp(escaped, "ig"), "<b>$&</b>");
        }
    }
}
